#include <sys/prctl.h>
#include <dlfcn.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>

#ifndef LDSO_ABI
#error "No LDSO_ABI defined"
#endif

#ifndef EXEC_ABI
#error "No EXEC_ABI defined"
#endif

#define PR_GET_FP_MODE 44
#define PR_FP_MODE_FR  (1 << 0)
#define PR_FP_MODE_FRE (1 << 1)

#define STR_VAL(VAL) #VAL
#define N_STR(VAL) STR_VAL(VAL)

#define START_STATE(NAME) 					\
case s_ ## NAME: 						\
  {								\
    switch (obj) 						\
      {

#define END_STATE						\
      default:							\
        return false;						\
      }								\
  break;							\
  }

#define NEXT(OBJ, NEXT_STATE)					\
case o_ ## OBJ: 						\
  current_fp_state = s_ ## NEXT_STATE;				\
  break;

#define NEXT_REQ_FR1(OBJ, NEXT_STATE)				\
case o_ ## OBJ:							\
  {								\
    if (has_fr1 || kernel_nofpmode)				\
      current_fp_state = s_ ## NEXT_STATE;			\
    else							\
      return false;						\
  }								\
  break;							

#define NEXT_REQ_FR0(OBJ, NEXT_STATE) 				\
case o_ ## OBJ:							\
  {								\
    if (!is_r6							\
        || (is_r6 && has_fr1 && has_fre))			\
      current_fp_state = s_ ## NEXT_STATE;			\
    else 							\
      return false;						\
  }								\
  break;

#define NEXT_REQ_FRE(OBJ, NEXT_STATE)				\
case o_ ## OBJ: 						\
  {								\
    if ((has_fr1 && has_fre) || kernel_nofpmode)		\
      current_fp_state = s_ ## NEXT_STATE;			\
    else							\
      return false;						\
  }								\
  break;

#define NEXT_NO_MODE_CHANGE(OBJ, NEXT_STATE)			\
case o_ ## OBJ: 						\
  {								\
    if (current_mode_valid (s_ ## NEXT_STATE))			\
      {								\
	current_fp_state = s_ ## NEXT_STATE;			\
	cant_change_mode = true;				\
      }								\
    else							\
      return false;						\
  }								\
  break;

static const char * const shared_lib_names[] =
{
  "libfpany.so", "libfpcorrupt.so", "libfpnone.so",
  "libfpext.so", "libfpsoft.so", "libfpsingle.so",
  "libfp32.so", "libfp64.so", "libfp64a.so",
  "libfpxx.so", "libfpxxo.so"
};

struct fp_mode_req
{
  int mode1;
  int mode2;
  int mode3;
};

enum fp_obj
{
  o_any=0,
  o_corrupt,
  o_none,
  o_ext,
  o_soft,
  o_single,
  o_fp32,
  o_fp64,
  o_fp64a,
  o_fpxx,
  o_fpxxo,
  o_max
};

enum fp_state
{
  s_any,
  s_none,
  s_soft,
  s_single,
  s_none_soft,
  s_none_single,
#if _MIPS_SIM == _ABIO32
  s_fp32,
  s_fpxx,
  s_fpxxo,
  s_fp64a,
  s_fpxxo_fpxx,
  s_fpxxo_fpxx_none,
  s_fp32_none,
  s_fp32_fpxx,
  s_fp32_fpxx_none,
  s_fp32_fpxxo,
  s_fp32_fpxxo_fpxx,
  s_fp32_fpxxo_none,
  s_fp32_fpxxo_fpxx_none,
  s_fp32_fp64a_fpxx,
  s_fp32_fp64a_fpxxo,
  s_fp32_fp64a_none,
  s_fp32_fp64a_fpxxo_fpxx,
  s_fp32_fp64a_fpxxo_none,
  s_fp32_fp64a_fpxx_none,
  s_fp32_fp64a_fpxxo_fpxx_none,
  s_none_fpxx,
  s_none_fpxxo,
  s_fp64a_fp32,
  s_fp64a_none,
  s_fp64a_fpxx,
  s_fp64a_fpxxo,
  s_fp64a_fpxx_none,
  s_fp64a_fpxxo_none,
  s_fp64a_fp64,
  s_fp64a_fp64_fpxx,
  s_fp64a_fp64_fpxxo,
  s_fp64a_fpxx_fpxxo,
  s_fp64a_fpxx_fpxxo_none,
  s_fp64a_fp64_fpxxo_fpxx,
  s_fp64_fpxx,
  s_fp64_fpxxo,
  s_fp64_fpxx_fpxxo,
#else
  s_none_fp64,
#endif
  s_fp64
};

static int current_fp_mode = -1;
static bool cant_change_mode = false;
static bool has_fr1 = false;
static bool has_fre = false;
static bool is_r6 = false;
static bool kernel_nofpmode = false;
static bool ldso_nofpmode = false;
static unsigned int fp_obj_count[o_max];
static enum fp_state current_fp_state = s_any;

static enum fp_obj
convert_to_fp_type (char *name, int *unload)
{

  *unload = 0;

  if (name[0] == '-')
    {
      *unload = 1;
      name++;
    }

  if (strcmp (name, "fp32") == 0)
    return o_fp32;
  else if (strcmp (name, "fpany") == 0)
    return o_any;
  else if (strcmp (name, "fpext") == 0)
    return o_ext;
  else if (strcmp (name, "fpnone") == 0)
    return o_none;
  else if (strcmp (name, "fpcorrupt") == 0)
    return o_corrupt;
  else if (strcmp (name, "fpsoft") == 0)
    return o_soft;
  else if (strcmp (name, "fpsingle") == 0)
    return o_single;
  else if (strcmp (name, "fp64") == 0)
    return o_fp64;
  else if (strcmp (name, "fp64a") == 0)
    return o_fp64a;
  else if (strcmp (name, "fpxx") == 0)
    return o_fpxx;
  else if (strcmp (name, "fpxxo") == 0)
    return o_fpxxo;
  else
    return o_max;
}

#ifdef _DEBUG
static void
print_obj (enum fp_obj obj)
{
  switch (obj)
    {
    case o_any:
      printf ("(any)");
      break;
    case o_corrupt:
      printf ("(corrupt)");
      break;
    case o_none:
      printf ("(none)");
      break;
    case o_ext:
      printf ("(ext)");
      break;
    case o_soft:
      printf ("(soft)");
      break;
    case o_single:
      printf ("(single)");
      break;
    case o_fp32:
      printf ("(fp32)");
      break;
    case o_fp64:
      printf ("(fp64)");
      break;
    case o_fp64a:
      printf ("(fp64a)");
      break;
    case o_fpxx:
      printf ("(fpxx)");
      break;
    case o_fpxxo:
      printf ("(fpxxo)");
      break;
    case o_max:
      break;
    }
}

static void
print_fp_state (enum fp_state state)
{
  switch (state)
    {
    case s_any:
      printf ("[any]");
      break;
    case s_none:
      printf ("[none]");
      break;
    case s_soft:
      printf ("[soft]");
      break;
    case s_single:
      printf ("[single]");
      break;
    case s_ext:
      printf ("[ext]");
      break;
    case s_none_soft:
      printf ("[none_soft]");
      break;
    case s_none_single:
      printf ("[none_single]");
      break;
#if _MIPS_SIM == _ABIO32
    case s_fp32:
      printf ("[fp32]");
      break;
    case s_fpxx:
      printf ("[fpxx]");
      break;
    case s_fpxxo:
      printf ("[fpxxo]");
      break;
    case s_fp64a:
      printf ("[fp64a]");
      break;
    case s_fpxxo_fpxx:
      printf ("[fpxxo_fpxx]");
      break;
    case s_fpxxo_fpxx_none:
      printf ("[fpxxo_fpxx_none]");
      break;
    case s_fp32_none:
      printf ("[fp32_none]");
      break;
    case s_fp32_fpxx:
      printf ("[fp32_fpxx]");
      break;
    case s_fp32_fpxx_none:
      printf ("[fp32_fpxx_none]");
      break;
    case s_fp32_fpxxo:
      printf ("[fp32_fpxxo]");
      break;
    case s_fp32_fpxxo_fpxx:
      printf ("[fp32_fpxxo_fpxx]");
      break;
    case s_fp32_fpxxo_fpxx_none:
      printf ("[fp32_fpxxo_fpxx_none]");
      break;
    case s_fp32_fp64a_fpxx:
      printf ("[fp32_fp64a_fpxx]");
      break;
    case s_fp32_fp64a_fpxxo:
      printf ("[fp32_fp64a_fpxxo]");
      break;
    case s_fp32_fp64a_none:
      printf ("[fp32_fp64a_none]");
      break;
    case s_fp32_fp64a_fpxxo_fpxx:
      printf ("[fp32_fp64a_fpxxo_fpxx]");
      break;
    case s_fp32_fp64a_fpxxo_none:
      printf ("[fp32_fp64a_fpxxo_none]");
      break;
    case s_fp32_fp64a_fpxx_none:
      printf ("[fp32_fp64a_fpxx_none]");
      break;
    case s_fp32_fp64a_fpxxo_fpxx_none:
      printf ("[fp32_fp64a_fpxxo_fpxx_none]");
      break;
    case s_none_fpxx:
      printf ("[none_fpxx]");
      break;
    case s_none_fpxxo:
      printf ("[none_fpxxo]");
      break;
    case s_fp64a_fp32:
      printf ("[fp64a_fp32]");
      break;
    case s_fp64a_none:
      printf ("[fp64a_none]");
      break;
    case s_fp64a_fpxx:
      printf ("[fp64a_fpxx]");
      break;
    case s_fp64a_fpxxo:
      printf ("[fp64a_fpxxo]");
      break;
    case s_fp64a_fpxx_none:
      printf ("[fp64a_fpxx_none]");
      break;
    case s_fp64a_fpxxo:
      printf ("[fp64a_fpxxo_none]");
      break;
    case s_fp64a_fp64:
      printf ("[fp64a_fp64]");
      break;
    case s_fp64a_fp64_fpxx:
      printf ("[fp64a_fp64_fpxx]");
      break;
    case s_fp64a_fp64_fpxxo:
      printf ("[fp64a_fp64_fpxxo]");
      break;
    case s_fp64a_fpxx_fpxxo:
      printf ("[fp64a_fpxx_fpxxo]");
      break;
    case s_fp64a_fpxx_fpxxo_none:
      printf ("[fp64a_fpxx_fpxxo_none]");
      break;
    case s_fp64a_fp64_fpxxo_fpxx:
      printf ("[fp64a_fp64_fpxxo_fpxx]");
      break;
    case s_fp64_fpxx:
      printf ("[fp64_fpxx]");
      break;
    case s_fp64_fpxxo:
      printf ("[fp64_fpxxo]");
      break;
    case s_fp64_fpxx_fpxxo:
      printf ("[fp64_fpxx_fpxxo]");
      break;
#else
    case s_none_fp64:
      printf ("[none_fp64]");
      break;
#endif
    case s_fp64:
      printf ("[fp64]");
      break;
    }
}
#endif

static bool
contains_fp64_or_fp64a (enum fp_state state)
{
#if _MIPS_SIM == _ABIO32
  switch (state)
    {
    case s_single:
    case s_fp32:
    case s_none:
    case s_none_fpxx:
    case s_none_fpxxo:
    case s_fp32_none:
    case s_fp32_fpxx:
    case s_fp32_fpxx_none:
    case s_fp32_fpxxo:
    case s_fp32_fpxxo_none:
    case s_fp32_fpxxo_fpxx:
    case s_fpxxo_fpxx_none:
    case s_fp32_fpxxo_fpxx_none:
    case s_fpxx:
    case s_fpxxo:
    case s_fpxxo_fpxx:
    case s_any:
    case s_soft:
    case s_none_soft:
    case s_none_single:
    default:
      return false;
    case s_fp64a:
    case s_fp64a_fpxx:
    case s_fp64a_fpxxo:
    case s_fp64a_fpxx_fpxxo:
    case s_fp64:
    case s_fp64_fpxx:
    case s_fp64_fpxxo:
    case s_fp64_fpxx_fpxxo:
    case s_fp64a_fp64:
    case s_fp64a_fp64_fpxx:
    case s_fp64a_fp64_fpxxo:
    case s_fp64a_fp64_fpxxo_fpxx:
    case s_fp64a_fp32:
    case s_fp64a_none:
    case s_fp64a_fpxx_none:
    case s_fp64a_fpxxo_none:
    case s_fp32_fp64a_none:
    case s_fp32_fp64a_fpxx:
    case s_fp32_fp64a_fpxxo:
    case s_fp32_fp64a_fpxxo_none:
    case s_fp32_fp64a_fpxx_none:
    case s_fp32_fp64a_fpxxo_fpxx:
    case s_fp32_fp64a_fpxxo_fpxx_none:
    case s_fp64a_fpxx_fpxxo_none:
      return true;
    }
#else
  return false;
#endif
}

/* This function will return the fp mode states for each valid fp mode.  */

static struct fp_mode_req
compute_fp_modes (enum fp_state state)
{
  struct fp_mode_req requirements;

  requirements.mode1 = -1;
  requirements.mode2 = -1;
  requirements.mode3 = -1;

#if _MIPS_SIM == _ABIO32
  switch (state)
    {
    case s_single:
      {
        if (is_r6)
	  requirements.mode1 = PR_FP_MODE_FR;
	else
	  {
	    requirements.mode1 = 0;
	    requirements.mode2 = PR_FP_MODE_FR;
	  }
	break;
      }
    case s_none:
    case s_none_fpxx:
    case s_none_fpxxo:
    case s_fpxxo_fpxx_none:
    case s_fp32:
    case s_fp32_none:
    case s_fp32_fpxx:
    case s_fp32_fpxx_none:
    case s_fp32_fpxxo:
    case s_fp32_fpxxo_none:
    case s_fp32_fpxxo_fpxx:
    case s_fp32_fpxxo_fpxx_none:
      {
	if (is_r6)
	  requirements.mode1 = PR_FP_MODE_FR | PR_FP_MODE_FRE;
	else
	  {
	    requirements.mode1 = 0;
	    requirements.mode2 = PR_FP_MODE_FR | PR_FP_MODE_FRE;
	  }
	break;
      }
    case s_fpxx:
    case s_fpxxo:
    case s_fpxxo_fpxx:
    case s_any:
    case s_soft:
    case s_none_soft:
    case s_none_single:
      {
	if (is_r6)
	  {
	    requirements.mode1 = PR_FP_MODE_FR;
	    requirements.mode2 = PR_FP_MODE_FR | PR_FP_MODE_FRE;
	  }
	else
	  {
	    requirements.mode1 = 0;
	    requirements.mode2 = PR_FP_MODE_FR;
	    requirements.mode3 = PR_FP_MODE_FR | PR_FP_MODE_FRE;
	  }
	break;
      }
    case s_fp64a:
    case s_fp64a_fpxx:
    case s_fp64a_fpxxo:
    case s_fp64a_fpxx_fpxxo:
      {
	requirements.mode1 = PR_FP_MODE_FR;
	requirements.mode2 = PR_FP_MODE_FR | PR_FP_MODE_FRE;
	break;
      }
    case s_fp64:
    case s_fp64_fpxx:
    case s_fp64_fpxxo:
    case s_fp64_fpxx_fpxxo:
    case s_fp64a_fp64:
    case s_fp64a_fp64_fpxx:
    case s_fp64a_fp64_fpxxo:
    case s_fp64a_fp64_fpxxo_fpxx:
      {
	requirements.mode1 = PR_FP_MODE_FR;
	break;
      }
    case s_fp64a_fp32:
    case s_fp64a_none:
    case s_fp64a_fpxx_none:
    case s_fp64a_fpxxo_none:
    case s_fp32_fp64a_none:
    case s_fp32_fp64a_fpxx:
    case s_fp32_fp64a_fpxxo:
    case s_fp32_fp64a_fpxxo_none:
    case s_fp32_fp64a_fpxx_none:
    case s_fp32_fp64a_fpxxo_fpxx:
    case s_fp32_fp64a_fpxxo_fpxx_none:
    case s_fp64a_fpxx_fpxxo_none:
      {
        requirements.mode1 = PR_FP_MODE_FR | PR_FP_MODE_FRE;
        break;
      }
    }
#else
  requirements.mode1 = PR_FP_MODE_FR;
#endif
  return requirements;
}

static bool
current_mode_valid (enum fp_state s)
{
  struct fp_mode_req req = compute_fp_modes (s);
  return (kernel_nofpmode || ldso_nofpmode
	  || req.mode1 == current_fp_mode
	  || req.mode2 == current_fp_mode
	  || req.mode3 == current_fp_mode);
}

static bool
set_next_fp_state (enum fp_obj obj)
{
  if (ldso_nofpmode
      && (contains_fp64_or_fp64a (current_fp_state)
	  || obj == o_fp64 || obj == o_fp64a))
    return false;

  switch (current_fp_state)
    {

    START_STATE(soft)
    NEXT(soft, soft)
    NEXT(any, soft)
    NEXT(none, none_soft)
    END_STATE

    START_STATE(none_soft)
    NEXT(none, none_soft)
    NEXT(soft, none_soft)
    NEXT(any, none_soft)
    END_STATE

    START_STATE(none_single)
    NEXT(none, none_single)
    NEXT(single, none_single)
    NEXT(any, none_single)
    END_STATE

    START_STATE(single)
    NEXT(single, single)
    NEXT(any, single)
    NEXT(none, none_single)
    END_STATE

#if _MIPS_SIM == _ABIO32
    START_STATE(any)
    NEXT_REQ_FR0(fp32, fp32)
    NEXT_REQ_FR0(none, none)
    NEXT(fpxx, fpxx)
    NEXT(fpxxo, fpxxo)
    NEXT_REQ_FR1(fp64a, fp64a)
    NEXT_REQ_FR1(fp64, fp64)
    NEXT(any,any)
    NEXT(soft,soft)
    NEXT(single,single)
    END_STATE

    START_STATE(none)
    NEXT(none, none)
    NEXT(any, none)
    NEXT(soft, none_soft)
    NEXT(single, none_single)
    NEXT(fp32, fp32_none)
    NEXT(fpxx, none_fpxx)
    NEXT(fpxxo, none_fpxxo)
    NEXT_REQ_FRE(fp64a, fp64a_none)
    END_STATE

    START_STATE(fp32)
    NEXT(fp32, fp32)
    NEXT(none, fp32_none)
    NEXT(fpxx, fp32_fpxx)
    NEXT(fpxxo, fp32_fpxxo)
    NEXT_REQ_FRE(fp64a, fp64a_fp32)
    NEXT(any, fp32)
    END_STATE

    START_STATE(fp32_none)
    NEXT(fp32, fp32_none)
    NEXT(none, fp32_none)
    NEXT(fpxx, fp32_fpxx_none)
    NEXT(fpxxo, fp32_fpxxo_none)
    NEXT_REQ_FRE(fp64a, fp32_fp64a_none)
    NEXT(any, fp32_none)
    END_STATE

    START_STATE(fpxx)
    NEXT_REQ_FR0(fp32, fp32_fpxx)
    NEXT_REQ_FR0(none, none_fpxx)
    NEXT_REQ_FR1(fp64, fp64_fpxx)
    NEXT_REQ_FR1(fp64a, fp64a_fpxx)
    NEXT(fpxxo, fpxxo_fpxx)
    NEXT(fpxx, fpxx)
    NEXT(any, fpxx)
    END_STATE

    START_STATE(fpxxo)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fpxxo)
    NEXT_NO_MODE_CHANGE(none, none_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64, fp64_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxx, fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fpxxo,fpxxo)
    NEXT_NO_MODE_CHANGE(any,fpxxo)
    END_STATE

    START_STATE(fp64a)
    NEXT_REQ_FRE(fp32, fp64a_fp32)
    NEXT_REQ_FRE(none, fp64a_none)
    NEXT(fp64, fp64a_fp64)
    NEXT(fpxxo, fp64a_fpxxo)
    NEXT(fpxx, fp64a_fpxx)
    NEXT(fp64a, fp64a)
    NEXT(any, fp64a)
    END_STATE

    START_STATE(fp64)
    NEXT(fp64a, fp64a_fp64)
    NEXT(fpxxo, fp64_fpxxo)
    NEXT(fpxx, fp64_fpxx)
    NEXT(fp64, fp64)
    NEXT(any, fp64)
    END_STATE

    START_STATE(fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(none, fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fp64, fp64_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxx, fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fpxxo, fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(any, fpxxo_fpxx)
    END_STATE

    START_STATE(fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(none, fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fpxxo, fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fpxx, fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fpxx_fpxxo_none)
    NEXT_NO_MODE_CHANGE(any, fpxxo_fpxx_none)
    END_STATE

    START_STATE(fp32_fpxx)
    NEXT(fp32, fp32_fpxx)
    NEXT(none, fp32_fpxx_none)
    NEXT(fpxx, fp32_fpxx)
    NEXT(fpxxo, fp32_fpxxo_fpxx)
    NEXT_REQ_FRE(fp64a, fp32_fp64a_fpxx)
    NEXT(any, fp32_fpxx)
    END_STATE

    START_STATE(fp32_fpxx_none)
    NEXT(fp32, fp32_fpxx_none)
    NEXT(none, fp32_fpxx_none)
    NEXT(fpxx, fp32_fpxx_none)
    NEXT(fpxxo, fp32_fpxxo_fpxx_none)
    NEXT_REQ_FRE(fp64a, fp32_fp64a_fpxx_none)
    NEXT(any, fp32_fpxx_none)
    END_STATE

    START_STATE(fp32_fpxxo)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fpxxo)
    NEXT_NO_MODE_CHANGE(none, fp32_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fpxxo, fp32_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxx, fp32_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fp64a, fp32_fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(any, fp32_fpxxo)
    END_STATE

    START_STATE(fp32_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(none, fp32_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fpxxo, fp32_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fpxx, fp32_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fp64a, fp32_fp64a_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(any, fp32_fpxxo_fpxx)
    END_STATE

    START_STATE(fp32_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fpxxo_none)
    NEXT_NO_MODE_CHANGE(none, fp32_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fpxxo, fp32_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fpxx, fp32_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp32_fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(any, fp32_fpxxo_none)
    END_STATE

    START_STATE(fp32_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(none, fp32_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fpxxo, fp32_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fpxx, fp32_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp32_fp64a_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(any, fp32_fpxxo_fpxx_none)
    END_STATE

    START_STATE(fp32_fp64a_fpxx)
    NEXT(fp32, fp32_fp64a_fpxx)
    NEXT(none, fp32_fp64a_fpxx_none)
    NEXT(fp64a, fp32_fp64a_fpxx)
    NEXT(fpxxo, fp32_fp64a_fpxxo_fpxx)
    NEXT(fpxx, fp32_fp64a_fpxx)
    NEXT(any, fp32_fp64a_fpxx)
    END_STATE

    START_STATE(fp32_fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(none, fp32_fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64a, fp32_fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxx, fp32_fp64a_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fpxxo, fp32_fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(any, fp32_fp64a_fpxxo)
    END_STATE

    START_STATE(fp32_fp64a_none)
    NEXT(fp32, fp32_fp64a_none)
    NEXT(none, fp32_fp64a_none)
    NEXT(fp64a, fp32_fp64a_none)
    NEXT(fpxxo, fp32_fp64a_fpxxo_none)
    NEXT(fpxx, fp32_fp64a_fpxx_none)
    NEXT(any, fp32_fp64a_none)
    END_STATE

    START_STATE(fp32_fp64a_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fp64a_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(none, fp32_fp64a_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp32_fp64a_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fpxx, fp32_fp64a_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fpxxo, fp32_fp64a_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(any, fp32_fp64a_fpxxo_fpxx)
    END_STATE

    START_STATE(fp32_fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(none, fp32_fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp32_fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fpxxo, fp32_fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fpxx, fp32_fp64a_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(any, fp32_fp64a_fpxxo_none)
    END_STATE

    START_STATE(fp32_fp64a_fpxx_none)
    NEXT(fp32, fp32_fp64a_fpxx_none)
    NEXT(none, fp32_fp64a_fpxx_none)
    NEXT(fp64a, fp32_fp64a_fpxx_none)
    NEXT(fpxx, fp32_fp64a_fpxx_none)
    NEXT(fpxxo, fp32_fp64a_fpxxo_fpxx_none)
    NEXT(any, fp32_fp64a_fpxx_none)
    END_STATE

    START_STATE(fp32_fp64a_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fp64a_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(none, fp32_fp64a_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp32_fp64a_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fpxx, fp32_fp64a_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fpxxo, fp32_fp64a_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(any, fp32_fp64a_fpxxo_fpxx_none)
    END_STATE

    START_STATE(none_fpxx)
    NEXT(none, none_fpxx)
    NEXT(fpxx, none_fpxx)
    NEXT(fp32, fp32_fpxx_none)
    NEXT(fpxxo, fpxxo_fpxx_none)
    NEXT_REQ_FRE(fp64a, fp64a_fpxx_none)
    NEXT(any, none_fpxx)
    END_STATE

    START_STATE(none_fpxxo)
    NEXT_NO_MODE_CHANGE(none, none_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxxo, none_fpxxo)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fpxx, fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(any, none_fpxxo)
    END_STATE

    START_STATE(fp64a_fp32)
    NEXT(fp32, fp64a_fp32)
    NEXT(fp64a, fp64a_fp32)
    NEXT(none, fp32_fp64a_none)
    NEXT(fpxxo, fp32_fp64a_fpxxo)
    NEXT(fpxx, fp32_fp64a_fpxx)
    NEXT(any, fp64a_fp32)
    END_STATE

    START_STATE(fp64a_none)
    NEXT(none, fp64a_none)
    NEXT(fp64a, fp64a_none)
    NEXT(fp32, fp32_fp64a_none)
    NEXT(fpxxo, fp64a_fpxxo_none)
    NEXT(fpxx, fp64a_fpxx_none)
    NEXT(any, fp64a_none)
    END_STATE

    START_STATE(fp64a_fpxx)
    NEXT_REQ_FRE(fp32, fp32_fp64a_fpxx)
    NEXT_REQ_FRE(none, fp64a_fpxx_none)
    NEXT(fp64a, fp64a_fpxx)
    NEXT(fpxx, fp64a_fpxx)
    NEXT(fpxxo, fp64a_fpxx_fpxxo)
    NEXT(fp64,fp64a_fp64_fpxx)
    NEXT(any, fp64a_fpxx)
    END_STATE

    START_STATE(fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(none, fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxx, fp64a_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxxo, fp64a_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64, fp64a_fp64_fpxxo)
    NEXT_NO_MODE_CHANGE(any, fp64a_fpxxo)
    END_STATE

    START_STATE(fp64a_fpxx_none)
    NEXT(fp32, fp32_fp64a_fpxx_none)
    NEXT(none, fp64a_fpxx_none)
    NEXT(fp64a, fp64a_fpxx_none)
    NEXT(fpxxo, fp64a_fpxx_fpxxo_none)
    NEXT(fpxx, fp64a_fpxx_none)
    NEXT(any, fp64a_fpxx_none)
    END_STATE

    START_STATE(fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(none, fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fpxx, fp64a_fpxx_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fpxxo, fp64a_fpxxo_none)
    NEXT_NO_MODE_CHANGE(any, fp64a_fpxxo_none)
    END_STATE

    START_STATE(fp64a_fp64)
    NEXT(fp64a, fp64a_fp64)
    NEXT(fpxxo, fp64a_fp64_fpxxo)
    NEXT(fpxx, fp64a_fp64_fpxx)
    NEXT(fp64, fp64a_fp64)
    NEXT(any, fp64a_fp64)
    END_STATE

    START_STATE(fp64a_fp64_fpxx)
    NEXT(fp64a, fp64a_fp64_fpxx)
    NEXT(fpxxo, fp64a_fp64_fpxxo_fpxx)
    NEXT(fpxx, fp64a_fp64_fpxx)
    NEXT(fp64, fp64a_fp64_fpxx)
    NEXT(any, fp64a_fp64_fpxx)
    END_STATE

    START_STATE(fp64a_fp64_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fp64_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxx, fp64a_fp64_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fpxxo, fp64a_fp64_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64, fp64a_fp64_fpxxo)
    NEXT_NO_MODE_CHANGE(any, fp64a_fp64_fpxxo)
    END_STATE

    START_STATE(fp64a_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fp64a_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(none, fp64a_fpxx_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxx, fp64a_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxxo, fp64a_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64,fp64a_fp64_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(any, fp64a_fpxx_fpxxo)
    END_STATE

    START_STATE(fp64a_fpxx_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fp32, fp32_fp64a_fpxxo_fpxx_none)
    NEXT_NO_MODE_CHANGE(none, fp64a_fpxx_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fpxx_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fpxx, fp64a_fpxx_fpxxo_none)
    NEXT_NO_MODE_CHANGE(fpxxo, fp64a_fpxx_fpxxo_none)
    NEXT_NO_MODE_CHANGE(any, fp64a_fpxx_fpxxo_none)
    END_STATE

    START_STATE(fp64a_fp64_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fp64_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fpxx, fp64a_fp64_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fpxxo, fp64a_fp64_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fp64, fp64a_fp64_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(any, fp64a_fp64_fpxxo_fpxx)
    END_STATE

    START_STATE(fp64_fpxx)
    NEXT(fp64a, fp64a_fp64_fpxx)
    NEXT(fpxxo, fp64_fpxx_fpxxo)
    NEXT(fpxx, fp64_fpxx)
    NEXT(fp64, fp64_fpxx)
    NEXT(any, fp64_fpxx)
    END_STATE

    START_STATE(fp64_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fp64_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxxo, fp64_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxx, fp64_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64, fp64_fpxxo)
    NEXT_NO_MODE_CHANGE(any, fp64_fpxxo)
    END_STATE

    START_STATE(fp64_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64a, fp64a_fp64_fpxxo_fpxx)
    NEXT_NO_MODE_CHANGE(fpxxo, fp64_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fpxx, fp64_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(fp64, fp64_fpxx_fpxxo)
    NEXT_NO_MODE_CHANGE(any, fp64_fpxx_fpxxo)
    END_STATE

#else

    START_STATE(any)
    NEXT(none, none)
    NEXT(fp64, fp64)
    NEXT(any,any)
    NEXT(soft,soft)
    NEXT(single,single)
    END_STATE

    START_STATE(fp64)
    NEXT(fp64, fp64)
    NEXT(none, none_fp64)
    NEXT(any, fp64)
    END_STATE

    START_STATE(none)
    NEXT(none, none)
    NEXT(any, none)
    NEXT(soft, none_soft)
    NEXT(single, none_single)
    NEXT(fp64, none_fp64)
    END_STATE

    START_STATE(none_fp64)
    NEXT(none, none_fp64)
    NEXT(any, none_fp64)
    NEXT(fp64, none_fp64)
    END_STATE

#endif
    }

  if (obj != o_max)
    fp_obj_count[obj]++;

  return true;
}

static int
remove_object (enum fp_obj obj)
{
  if (obj == o_max)
    return false;

  fp_obj_count[obj]--;

  /* We can't change fp state until all the objects
     of a particular type have been unloaded.  */
  if (fp_obj_count[obj] != 0)
    return false;

  switch (current_fp_state)
    {
    START_STATE(soft)
    NEXT(soft,any)
    END_STATE

    START_STATE(none)
    NEXT(none,any)
    END_STATE

    START_STATE(single)
    NEXT(single,any)
    END_STATE

    START_STATE(any)
    NEXT(any,any)
    END_STATE

    START_STATE(fp64)
    NEXT(fp64, any)
    END_STATE

    START_STATE(none_soft)
    NEXT(soft,none)
    NEXT(none,soft)
    END_STATE

    START_STATE(none_single)
    NEXT(single,none)
    NEXT(none,single)
    END_STATE

#if _MIPS_SIM == _ABIO32
    START_STATE(fp32)
    NEXT (fp32,any)
    END_STATE

    START_STATE(fpxx)
    NEXT (fpxx,any)
    END_STATE

    START_STATE(fpxxo)
    NEXT (fpxxo,any)
    END_STATE

    START_STATE(fp64a)
    NEXT(fp64a, any)
    END_STATE

    START_STATE(fpxxo_fpxx)
    NEXT(fpxx, fpxxo)
    NEXT(fpxxo, fpxx)
    END_STATE

    START_STATE(fpxxo_fpxx_none)
    NEXT(fpxx, none_fpxxo)
    NEXT(fpxxo, none_fpxx)
    NEXT(none, fpxxo_fpxx)
    END_STATE

    START_STATE(fp32_none)
    NEXT(fp32, none)
    NEXT(none, fp32)
    END_STATE

    START_STATE(fp32_fpxx)
    NEXT(fp32, fpxx)
    NEXT(fpxx, fp32)
    END_STATE

    START_STATE(fp32_fpxx_none)
    NEXT(fp32, none_fpxx)
    NEXT(fpxx, fp32_none)
    NEXT(none, fp32_fpxx)
    END_STATE

    START_STATE(fp32_fpxxo)
    NEXT(fp32, fpxxo)
    NEXT(fpxxo, fp32)
    END_STATE

    START_STATE(fp32_fpxxo_fpxx)
    NEXT(fp32, fpxxo_fpxx)
    NEXT(fpxxo, fp32_fpxx)
    NEXT(fpxx, fp32_fpxxo)
    END_STATE

    START_STATE(fp32_fpxxo_none)
    NEXT(fp32, none_fpxxo)
    NEXT(fpxxo, fp32_none)
    NEXT(none, fp32_fpxxo)
    END_STATE

    START_STATE(fp32_fpxxo_fpxx_none)
    NEXT(fp32, fpxxo_fpxx_none)
    NEXT(fpxxo, fp32_fpxx_none)
    NEXT(fpxx, fp32_fpxxo_none)
    NEXT(none, fp32_fpxxo_fpxx)
    END_STATE

    START_STATE(fp32_fp64a_fpxx)
    NEXT(fp32, fp64a_fpxx)
    NEXT(fp64a, fp32_fpxx)
    NEXT(fpxx, fp64a_fp32)
    END_STATE

    START_STATE(fp32_fp64a_fpxxo)
    NEXT(fp32, fp64a_fpxxo)
    NEXT(fp64a, fp32_fpxxo)
    NEXT(fpxxo, fp64a_fp32)
    END_STATE

    START_STATE(fp32_fp64a_none)
    NEXT(fp32, fp64a_none)
    NEXT(fp64a, fp32_none)
    NEXT(none, fp64a_fp32)
    END_STATE

    START_STATE(fp32_fp64a_fpxxo_fpxx)
    NEXT(fp32, fp64a_fpxx_fpxxo)
    NEXT(fp64a, fp32_fpxxo_fpxx)
    NEXT(fpxx, fp32_fp64a_fpxxo)
    NEXT(fpxxo, fp32_fp64a_fpxx)
    END_STATE

    START_STATE(fp32_fp64a_fpxxo_none)
    NEXT(fp32, fp64a_fpxxo_none)
    NEXT(fp64a, fp32_fpxxo_none)
    NEXT(fpxxo, fp32_fp64a_none)
    NEXT(none, fp32_fp64a_fpxxo)
    END_STATE

    START_STATE(fp32_fp64a_fpxx_none)
    NEXT(fp32, fp64a_fpxx_none)
    NEXT(fp64a, fp32_fpxx_none)
    NEXT(fpxx, fp32_fp64a_none)
    NEXT(none, fp32_fp64a_fpxx)
    END_STATE

    START_STATE(fp32_fp64a_fpxxo_fpxx_none)
    NEXT(fp32, fp64a_fpxx_fpxxo_none)
    NEXT(fp64a, fp32_fpxxo_fpxx_none)
    NEXT(fpxxo, fp32_fp64a_fpxx_none)
    NEXT(fpxx, fp32_fp64a_fpxxo_none)
    NEXT(none, fp32_fp64a_fpxxo_fpxx)
    END_STATE

    START_STATE(none_fpxx)
    NEXT(none, fpxx)
    NEXT(fpxx, none)
    END_STATE

    START_STATE(none_fpxxo)
    NEXT(none, fpxxo)
    NEXT(fpxxo, none)
    END_STATE

    START_STATE(fp64a_fp32)
    NEXT(fp32, fp64a)
    NEXT(fp64a, fp32)
    END_STATE

    START_STATE(fp64a_none)
    NEXT(none, fp64a)
    NEXT(fp64a, none)
    END_STATE

    START_STATE(fp64a_fpxx)
    NEXT(fp64a, fpxx)
    NEXT(fpxx, fp64a)
    END_STATE

    START_STATE(fp64a_fpxxo)
    NEXT(fp64a, fpxxo)
    NEXT(fpxxo, fp64a)
    END_STATE

    START_STATE(fp64a_fpxx_none)
    NEXT(fp64a, none_fpxx)
    NEXT(fpxx, fp64a_none)
    NEXT(none, fp64a_fpxx)
    END_STATE

    START_STATE(fp64a_fpxxo_none)
    NEXT(fp64a, none_fpxxo)
    NEXT(fpxxo, fp64a_none)
    NEXT(none, fp64a_fpxxo)
    END_STATE

    START_STATE(fp64a_fp64)
    NEXT(fp64a, fp64)
    NEXT(fp64, fp64a)
    END_STATE

    START_STATE(fp64a_fp64_fpxx)
    NEXT(fp64a, fp64_fpxx)
    NEXT(fpxx, fp64a_fp64)
    NEXT(fp64, fp64a_fpxx)
    END_STATE

    START_STATE(fp64a_fp64_fpxxo)
    NEXT(fp64a, fp64_fpxxo)
    NEXT(fpxxo, fp64a_fp64)
    NEXT(fp64, fp64a_fpxxo)
    END_STATE

    START_STATE(fp64a_fpxx_fpxxo)
    NEXT(fp64a, fpxxo_fpxx)
    NEXT(fpxx, fp64a_fpxxo)
    NEXT(fpxxo, fp64a_fpxx)
    END_STATE

    START_STATE(fp64a_fpxx_fpxxo_none)
    NEXT(fp64a, fpxxo_fpxx_none)
    NEXT(fpxx, fp64a_fpxxo_none)
    NEXT(fpxxo, fp64a_fpxx_none)
    NEXT(none, fp64a_fpxx_fpxxo)
    END_STATE

    START_STATE(fp64a_fp64_fpxxo_fpxx)
    NEXT(fp64a, fp64_fpxx_fpxxo)
    NEXT(fpxx, fp64a_fp64_fpxxo)
    NEXT(fpxxo, fp64a_fp64_fpxx)
    NEXT(fp64, fp64a_fpxx_fpxxo)
    END_STATE

    START_STATE(fp64_fpxx)
    NEXT(fpxx, fp64)
    NEXT(fp64, fpxx)
    END_STATE

    START_STATE(fp64_fpxxo)
    NEXT(fpxxo, fp64)
    NEXT(fp64, fpxxo)
    END_STATE

    START_STATE(fp64_fpxx_fpxxo)
    NEXT(fp64, fpxxo_fpxx)
    NEXT(fpxxo, fp64_fpxx)
    NEXT(fpxx, fp64_fpxxo)
    END_STATE

#else

    START_STATE(none_fp64)
    NEXT(none, fp64)
    NEXT(fp64, none)
    END_STATE

#endif
    }

    return true;
}

static bool
is_fp_mode_valid (void)
{
  int prev_fp_mode;
  bool match = false;

  /* Get the current fp mode.  */
  prev_fp_mode = current_fp_mode;

  current_fp_mode = prctl (PR_GET_FP_MODE);

  /* If the prctl call fails assume the core only has FR0 mode support.  */
#if _MIPS_SIM == _ABIO32
  if (current_fp_mode == -1 || ldso_nofpmode)
    current_fp_mode = 0;
#else
  if (current_fp_mode == -1)
    current_fp_mode = PR_FP_MODE_FR;
#endif

  match = current_mode_valid (current_fp_state);

  if (!match)
    return false;

  /* Check if mode changes are not allowed but a mode change happened.  */
  if (cant_change_mode && current_fp_mode != prev_fp_mode)
    return false;

  return true;
}

int
main (int argc, char* argv[])
{
  void * shared_lib_ptrs[o_max];
  int i;
  enum fp_obj obj;
  int unload;
  int success;
  int apos = 1;
  enum fp_obj ldso_obj = convert_to_fp_type (N_STR(LDSO_ABI), &unload);
  enum fp_obj exec_obj = convert_to_fp_type (N_STR(EXEC_ABI), &unload);

  for (i = 0; i < o_max ; i++)
    {
      fp_obj_count[i] = 0;
      shared_lib_ptrs[i] = 0;
    }

  /* Look for any arguments describing the possible hardware modes.  */
  for (apos = 1; apos < argc; apos++)
    {
      if (argv[apos][0] != 'm')
	break;
      if (strcmp ("mfr1", argv[apos]) == 0)
        has_fr1 = true;
      else if (strcmp ("mfre", argv[apos]) == 0)
	has_fre = true;
      else if (strcmp ("mnofpmode", argv[apos]) == 0)
	kernel_nofpmode = true;
      else if (strcmp ("mr6", argv[apos]) == 0)
	{
	  is_r6 = true;
	  has_fr1 = true;
	}
    }

  current_fp_mode = prctl (PR_GET_FP_MODE);

  /* If the prctl call fails assume the core only has FR0 mode support.  */
#if _MIPS_SIM == _ABIO32
  if (current_fp_mode == -1 || kernel_nofpmode)
    current_fp_mode = 0;
#else
  if (current_fp_mode == -1)
    current_fp_mode = PR_FP_MODE_FR;
#endif

  if (!kernel_nofpmode)
  {
    if ((is_r6 && !(current_fp_mode & PR_FP_MODE_FR))
	|| (!has_fr1 && (current_fp_mode & PR_FP_MODE_FR))
	|| (!has_fre && (current_fp_mode & PR_FP_MODE_FRE)))
      return 3;
  }

  /* Set the initial fp state by using the dynamic linker executable abis.  */
  cant_change_mode = false;

  if (ldso_obj != o_none)
    {
      if (!set_next_fp_state (ldso_obj))
	return 4;
    }

  if (!set_next_fp_state (exec_obj))
    return 5;

  if (ldso_obj == o_none)
    {
      if (!set_next_fp_state (ldso_obj))
	return 4;
    }

  if (kernel_nofpmode)
    {
      kernel_nofpmode = false;
      ldso_nofpmode = true;
    }

  if (!ldso_nofpmode && !is_fp_mode_valid ())
    return 6;

  /* Loop over the arguments.  */
  for (; apos < argc; apos++)
    {
      /* Convert the object string to the correct fp type number.  */
      obj = convert_to_fp_type (argv[apos], &unload);

      if (obj == o_max)
        return 7;

      cant_change_mode = false;

      /* Perform the load/unload.  */
      if (unload)
	{
	  remove_object (obj);
	  if (dlclose (shared_lib_ptrs[obj]) != 0)
	    return 8;

	  shared_lib_ptrs[obj] = 0;
	}
      else
	{
	  success = set_next_fp_state (obj);
          shared_lib_ptrs[obj] = dlopen (shared_lib_names[obj], RTLD_LAZY);

	  /* If we expected an error and the load was successful then fail.  */
          if (!success && (shared_lib_ptrs[obj] != 0))
	    return 9;
          if (success && (shared_lib_ptrs[obj] == 0))
	    return 10;
	}

      if (!ldso_nofpmode && !is_fp_mode_valid ())
	return 11;

    }

  return 0;
}
