FAIL_LOAD="abitest-fp64-fpcorrupt abitest-fp64-fpext abitest-fpany-fpcorrupt \
           abitest-fpcorrupt-fp64 abitest-fpcorrupt-fpany abitest-fpcorrupt-fpcorrupt \
           abitest-fpcorrupt-fpext abitest-fpcorrupt-fpnone abitest-fpcorrupt-fpsingle \
           abitest-fpcorrupt-fpsoft abitest-fpext-fp64 abitest-fpext-fpcorrupt \
           abitest-fpext-fpnone abitest-fpext-fpsingle abitest-fpext-fpsoft \
           abitest-fpnone-fpcorrupt abitest-fpnone-fpext abitest-fpsingle-fpcorrupt \
           abitest-fpsingle-fpext abitest-fpsoft-fpcorrupt abitest-fpsoft-fpext \
           abitest-fp64-fpsingle abitest-fp64-fpsoft abitest-fpsingle-fpsoft \
           abitest-fpsingle-fp64 abitest-fpany-fpext abitest-fpext-fpany \
           abitest-fpext-fpext abitest-fpsoft-fp64 abitest-fpsoft-fpsingle"

REQUIRE_FR1="abitest-fp64-fp64 abitest-fp64-fpany abitest-fp64-fpnone \
             abitest-fpany-fp64 abitest-fpany-fpany abitest-fpany-fpnone \
             abitest-fpany-fpsingle abitest-fpany-fpsoft abitest-fpnone-fp64 \
             abitest-fpnone-fpany abitest-fpnone-fpnone abitest-fpsingle-fpany \
             abitest-fpsingle-fpsingle abitest-fpsoft-fpany abitest-fpsoft-fpsoft \
             abitest-fpsingle-fpnone abitest-fpnone-fpsingle \
             abitest-fpnone-fpsoft abitest-fpsoft-fpnone"


PASSING_TESTS="$REQUIRE_FR1"

FP_ABIS="fpext fpnone fpcorrupt fpsoft fpsingle fp64"
CPU=
fail=0
core_supports_r6=0
core_supports_fr1=0
core_supports_fre=0

if [ $# -eq 1 -a $1 == "r5sim"  ]; then
  echo "MIPS64R2-generic"
  ./$0 cpu=MIPS64R2-generic fr1
  if [ $? -ne 0 ]; then
    fail=1
  fi
  if [ $fail -eq 1 ]; then
    echo "FAILED"
    exit 1
  fi
  exit 0
else
  while [ $# != 0 ]; do
    case $1 in
      cpu=*) CPU=${1##cpu=};;
      r6) core_supports_r6=1;;
      fr1) core_supports_fr1=1;;
      fre) core_supports_fre=1;;
      *) echo "unknown argument $1"; exit 1;;
    esac
    shift
  done
fi

# Set up the CPU and supported modes
if [ "x$CPU" == "x" ]; then
  SIM=
  export LD_LIBRARY_PATH=.
else
  SIM="/fixme/bin/qemu-mips64el -cpu $CPU -E LD_LIBRARY_PATH=."
fi

FLAGS=""
if [ $core_supports_r6 -eq 1 ]; then
  core_supports_fr1=1
  FLAGS="$FLAGS mr6"
fi
if [ $core_supports_fr1 -eq 1 ]; then
  FLAGS="$FLAGS mfr1"
fi
if [ $core_supports_fre -eq 1 ]; then
  FLAGS="$FLAGS mfre"
fi
testcount=0

# A test that should be successfully loaded onto the current CPU and
# which should therefore pass regardless of what extra libraries
# are given to it. This is because the test program is designed
# to verify that libraries are correctly loaded or rejected
# appropriately and never fail if the dynamic loader behaves
# correctly.
# This is testing the dynamic loader.

function expect_pass_helper
{
  echo -n "."
  ((testcount++))
  testname=$1
  shift
  $SIM $testname $FLAGS ${@} 2> /dev/null > /dev/null
  exit_code=$?
  if [ $exit_code -ne 0 ]; then
    echo "'$testname ${@}' should work. Received $exit_code"
    fail=1
  fi
}

function expect_pass
{
  for a in $FP_ABIS ; do
    expect_pass_helper $1 $a fpany
    expect_pass_helper $1 fpany $a
  done

  m1list=( fpnone fp64 fpsoft )
  for m1 in ${m1list[@]}; do
    m2list=( ${m1list[@]%%$m1} )
    for m2 in ${m2list[@]}; do
      expect_pass_helper $1 $m1 $m2 ${m2list[@]%%$m2}
    done
  done

}

# A test which is checking that the program loader refuses to load a program
# which cannot possibly execute in any supported hardware mode for the
# current CPU.
# This is testing the program loader

function expect_fail
{
  echo -n "@"
  ((testcount++))
  $SIM $1 $FLAGS 2> /dev/null > /dev/null
  exit_code=$?
  if [ $exit_code -ne 137 ]; then
    echo "'$1' should be rejected by loader with code 137. Received $exit_code"
    fail=1
  fi
}

for i in $FAIL_LOAD ; do
  expect_fail $i
done

for i in $REQUIRE_FR1 ; do
  if [ $core_supports_fr1 -eq 1 ]; then
    # R6 or R<6 with FR1
    expect_pass $i
  else
    echo "Broken config, all N64 ABIs require FR=1"
    fail=1
  fi
done

echo ""
echo "testcount: $testcount"
exit $fail
